#!/bin/ksh93
#set tabstop=4 to display correctly in vi
#
#-------------------------------------------------------------------------------
# File Name:    ArchiveLog.ksh (ksh93)
#				Based on existing copyleft code at website http://www.sustainablesoftware.net - please respect these terms.
# Author:       Alex Levy t/a http://www.sustainablesoftware.net
#
# Purpose:      Archive the current log file periodically through a scheduler.
#               This may be necessary to meet SLAs for recovery on databases with large log files but minimal update activity.
# 
# Usage:        ArchiveLog.ksh instance_name database_name 
#               Assumes, but does not enforce, runner is the instance owner.
#
# File dependencies:
#               ${INSTHOME}/sqllib/db2profile
#               ${INSTHOME}/sqllib/userprofile 
#               ${INSTHOME}/sqllib/db2nodes.cfg 
#               ${INSTHOME}/FUNCTIONS   (the userprofile exports $FPATH).
#				${INSTHOME}/MAINTENANCE (the userprofile exports $MAINTENANCE_DIR).
#				${INSTHOME}/MAINTENANCE/log  (the userprofile exports $MAINTENANCE_LOGS).
#				${INSTHOME}/MAINTENANCE/log/work (the userprofile exports $MAINTENANCE_WORK).
#
# Calls:		${MAINTENANCE}/Exit.ksh on error	
#
# File dependencies for partitioned databases:
#               None - the working assumption is that $INSTHOME is network-mounted on DPF systems.
#               The command run - "archive log for db $1 on all dbpartitionnums" - is also effective on single-partition instances.
#
# Notes:        In general external backup/recovery solutions such as TSM manage the expiry of backup images and archive logs at larger sites.
#         		This script does not prune database backup images or archive logs;
#               if historic backup images and/or archive logs are retained locally on disk, then you can backup and prune them in step using:
#					http://www.sustainablesoftware.net/downloads/BackupArchivePrune.ksh 
#
# Date:         April 2012.           
#-------------------------------------------------------------------------------

# Functions in $FPATH never exit, they only return.
# We can either test each return code or use a trap. If we do use a trap, and we wish to test the return code first,
# then the test must be part of the same statement, e.g. F_CONNECT $DBNAME $DEBUG_MODE || die.
# The trap statement and the 2 inline functions 'FatalError' and 'die' handle the exit for all error conditions.
# The 'die' function is called on any error detected by the program logic.
# The FatalError function is triggered by any non-zero signal, including the exit code from 'die'.

trap FatalError ERR

function FatalError {
# Single point of exit on any error condition.
# This relies on a signal trap. ERR is any non-zero exit status and is Korn shell only. 
# See 'man trap' for further detail.
# Can also call this function by 'kill -s ERR $$ 2>/dev/null'

RC=$? 
if   (( RC == 99 ))
then # error has already been trapped by function 'die', do nothing more except exit
	 ${MAINTENANCE_DIR}/Exit.ksh 99	
else
     print ERR signal received
	 if   [[ $LOG == "NOLOG" || ! -f $LOG || ! -s $LOG ]]
	 then F_MESSAGE "$script Ends with error(s), see stdout above." PLAIN I +x
	 else F_RESET_OUTPUT
     	  cat $LOG
     	  F_MESSAGE "$script ends with error(s). Please review log file $LOG . " PLAIN I +x | tee -a $LOG
     	  F_MESSAGE "The log file is displayed above for ease of reference. " PLAIN I +x | tee -a $LOG
	 fi
	 ${MAINTENANCE_DIR}/Exit.ksh 4	
fi
}

function die {
# Single point of exit on any error condition caught by the program logic.
# Restore stdout and stderr if required, before flowing to FatalError.

LOG=${LOG:=NOLOG} # use $LOG if set, otherwise assign a value of NOLOG to $LOG.
LINE=${1:=unknown}
LOCATION=${2:=unknown}

if   [[ $LOG == "NOLOG" || ! -f $LOG || ! -s $LOG ]]
then F_MESSAGE "$script Ends with error(s) at line $LINE, location $LOCATION; see stdout above." PLAIN I +x
else F_RESET_OUTPUT
     cat $LOG
     F_MESSAGE "$script ends with error(s). Please review log file $LOG . " PLAIN I +x | tee -a $LOG
     F_MESSAGE "The log file is displayed above for ease of reference. " PLAIN I +x | tee -a $LOG
fi

# trigger FatalError
exit 99
}

#-------------------------------------------------------------------------------
# Declare and test variables.
#-------------------------------------------------------------------------------

set -u								# assert all variables are defined and set.	 
typeset script=$(basename $0 .ksh)  # Strip off path and '.ksh' extension.
DEBUG_MODE="+x"                     # Default value - disable debugging.
typeset -i RC=0						# return code from function calls and o/s
typeset -i count=999				
now=$(date +%Y%m%d%H%M%S)              
typeset -uL3 LOGARCHMETH1="???"    	# reset later
typeset LOG=NOLOG               	# reset later

# Check the call
(( $# == 2 )) || F_ERROR_MESSAGE " Usage! I said:  $0 <instance_name> <database_name>" $DEBUG_MODE   
(( $# == 2 )) || die $LINENO 1 
typeset -l  INSTNAME=$1
typeset -u  DBNAME=$2
typeset     INSTHOME=~${INSTNAME}

# Assert all required files and directories are in place and are accessible or executable;
# the first function call tests for $INSTHOME and the db2profile and executes db2profile.
F_SETUP_DB2_ENVIRONMENT $INSTHOME $INSTNAME $DBNAME $DEBUG_MODE || die  $LINENO 2
F_DIR_CHECK $MAINTENANCE_DIR  $DEBUG_MODE   || die $LINENO 3
F_DIR_CHECK $MAINTENANCE_LOGS $DEBUG_MODE   || die $LINENO 4
F_DIR_CHECK $MAINTENANCE_WORK $DEBUG_MODE   || die $LINENO 5 

# check the database argument (is it catalogued as a local database?)
F_TEST_DBNAME $DBNAME $DEBUG_MODE || die $LINENO 6

# With these basic checks out of the way, direct all output to a timestamped log file.
LOG=${MAINTENANCE_LOGS}/${script}.${DBNAME}.${now}.log
F_INIT_OUTPUT_FILE $INSTNAME $LOG $DEBUG_MODE || die $LINENO 7
export LOG				 
F_REDIRECT_OUTPUT $LOG $DEBUG_MODE || die $LINENO 8

export DB2NODE=0 # only meaningful in DPF

# check no backup is running before connecting
db2 -v attach to $INSTNAME || die $LINENO 9
while (( count > 0 ))
do
    count=$(db2 list utilities | grep -ic backup | grep $DBNAME ) || : #suppress Unix trap of return code 1 if no match found
    (( count > 0 )) && F_MESSAGE "Backup(s) are running." PLAIN I $DEBUG_MODE
    (( count > 0 )) && sleep 300
done

 
# Disable trap from here on, since we want an explicit return code check for the archive log command. 
trap " " ERR

# Obvious, but SQL2417N  Archive log is not allowed on non-recoverable databases. 
# The db cfg LOGARCHMETH1 can contain off|logretain|userexit|tsm|vendor, so we test for 'off'.
F_CONNECT $DBNAME $DEBUG_MODE || die $LINENO 10
F_RUN_SQL "select value from  sysibmadm.dbcfg where name = 'logarchmeth1'" "-xnq" 0 $DEBUG_MODE
RC=$?; F_TEST_DB2_RETURNCODE $RC 0 $0 11 "Error querying logarchmeth1 value from sysibmadm.dbcfg " || die $LINENO 11
LOGARCHMETH1=${SQLVAR1}

# The archive log command establishes its own connection.
F_TERMINATE $DEBUG_MODE  

if  [[ ${LOGARCHMETH1} == "OFF" ]]
then
    F_MESSAGE "$0 : Database $DBNAME is not configured for archive logging, so no logs have been archived! Bye!" PLAIN I $DEBUG_MODE
else
   # Disable trap from here on, since we want an explicit return code check for the archive log command. 
   trap " " ERR
   F_RUN_DB2CLP "archive log for db $DBNAME on all dbpartitionnums" 0 ${MAINTENANCE_WORK}/${script}.archive_log nocompress noattach $DEBUG_MODE	
   RC=$? ; F_TEST_DB2_RETURNCODE $RC 0 $0 11 "Archive log ommand error." || die $LINENO 11
   cat  ${MAINTENANCE_WORK}/${script}.archive_log
fi

# Restore stdout and stderr to terminal
F_RESET_OUTPUT 
cat $LOG
exit 0
