#!/bin/ksh93
#set tabstop=4 to display correctly in vi
#
#-------------------------------------------------------------------------------
# File Name:    Check_Integrity.ksh (ksh93)
#				Based on existing copyleft code at website http://www.sustainablesoftware.net - please respect these terms.
#
# Purpose:      Check referential integrity for application tables on systems where RI is declared but not enforced
#               - for example data warehousing environments and SQL Replication apply servers.
#
# Usage:        Check_Integrity.ksh $1 instance_name $2 database_name 
#               Enforces assumption that the runner is the instance owner.
# 
# Out of scope: check constraints
#
# File dependencies:
#               ${INSTHOME}/sqllib/db2profile
#               ${INSTHOME}/sqllib/userprofile 
#               ${INSTHOME}/FUNCTIONS            - the userprofile exports this as $FPATH.
#				${INSTHOME}/MAINTENANCE          - the userprofile exports this as $MAINTENANCE_DIR.
#				${INSTHOME}/MAINTENANCE/log      - the userprofile exports this as $MAINTENANCE_LOGS.
#				${INSTHOME}/MAINTENANCE/log/work - the userprofile exports this as $MAINTENANCE_WORK.
#				${INSTHOME}/MAINTENANCE/generate_integrity_checks.sql (sic)
#
# File dependencies for partitioned databases:
#               None - this script operates on all partitions.
#
# Calls:		${MAINTENANCE}/Exit.ksh on error	
#
# Notes:        The .sql component checks each FK/PK column pairing individually.
#				That is, a PK of (col1,col2) will only generate 2 SQL statements:
#				to check all child col1 values are in the parent column col1 and all child col2 values are in parent column col2.
#				This is probably good enough for most purposes, given the modern fashion for single-column identity or surrogate keys.
#				Extra code could be built-in to cater for composite primary keys, at the expense of some complexity.
#
# Date:         February 2010
# Modified      11.06.2012 Re-integrate into a complete maintenance suite for DB2 9.7  
#
# Author:       Alex Levy t/a http://www.sustainablesoftware.net
# Credits:		
# Date:         June 2012.           
#-------------------------------------------------------------------------------

# Functions in $FPATH never exit, they only return.
# We can either test each return code or use a trap. If we do use a trap, and we wish to test the return code first,
# then the test must be part of the same statement, e.g. F_CONNECT $DBNAME $DEBUG_MODE || die.
# The trap statement and the 2 inline functions 'FatalError' and 'die' can handle the exit for all error conditions.
# The 'die' function is called on any error detected by the program logic.
# The FatalError function is triggered by any non-zero signal, including the exit code from 'die'.

trap FatalError ERR

function FatalError {
# Single point of exit on any error condition.
# This relies on a signal trap. ERR is any non-zero exit status and is Korn shell only. 
# See 'man trap' for further detail.
# Can also call this function by 'kill -s ERR $$ 2>/dev/null'

RC=$? 
if   (( RC == 99 ))
then # error has already been trapped by function 'die', do nothing more except exit
	 ${MAINTENANCE_DIR}/Exit.ksh 99	
else
     print ERR signal received
	 if   [[ $LOG == "NOLOG" || ! -f $LOG || ! -s $LOG ]]
	 then F_MESSAGE "$script Ends with error(s), see stdout above." PLAIN I +x
	 else F_RESET_OUTPUT
     	  cat $LOG
     	  F_MESSAGE "$script ends with error(s). Please review log file $LOG . " PLAIN I +x | tee -a $LOG
     	  F_MESSAGE "The log file is displayed above for ease of reference. " PLAIN I +x | tee -a $LOG
	 fi
	 ${MAINTENANCE_DIR}/Exit.ksh 4	
fi
}

function die {
# Single point of exit on any error condition caught by the program logic.
# Restore stdout and stderr if required, before flowing to FatalError.

LOG=${LOG:=NOLOG} # use $LOG if set, otherwise assign a value of NOLOG to $LOG.
LINE=${1:=unknown}
LOCATION=${2:=unknown}

if   [[ $LOG == "NOLOG" || ! -f $LOG || ! -s $LOG ]]
then F_MESSAGE "$script Ends with error(s) at line $LINE, location $LOCATION; see stdout above." PLAIN I +x
else F_RESET_OUTPUT
     cat $LOG
     F_MESSAGE "$script ends with error(s) at line $LINE, location ${LOCATION}. Please review log file $LOG . " PLAIN I +x | tee -a $LOG
     F_MESSAGE "The log file is displayed above for ease of reference. " PLAIN I +x | tee -a $LOG
fi

# trigger FatalError
exit 99
} 

#-------------------------------------------------------------------------------
# Declare and test variables.
#-------------------------------------------------------------------------------

set -u								# assert all variables are defined and set.	 
typeset script=$(basename $0 .ksh)  # Strip off path and '.ksh' extension.
DEBUG_MODE="+x"                     # Default value - disable debugging.
typeset -i RC=0						# return code from function calls and o/s
typeset -i counter=999			
now=$(date +%Y%m%d%H%M%S)              
typeset LOG=NOLOG               	# set later

# Check the call
(( $# == 2 )) || F_ERROR_MESSAGE " Usage! I said:  $0 <instance_name> <database_name>" $DEBUG_MODE   
(( $# == 2 )) || die $LINENO 1 
typeset -l  INSTNAME=$1
typeset -l  DBNAME=$2
if   [[ $INSTNAME == $USER ]]
then :
else
     F_ERROR_MESSAGE "$0 : The runner must be the DB2 instance owner. Exiting ..." $DEBUG_MODE
     die $LINENO 2
fi

# Assert all required files and directories are in place and are accessible or executable;
# the first function call tests for $INSTHOME and the db2profile and executes db2profile.
F_SETUP_DB2_ENVIRONMENT $INSTHOME $INSTNAME $DBNAME $DEBUG_MODE || die $LINENO 3
for dir in 	$FPATH 					\
			$MAINTENANCE_DIR 		\
			$MAINTENANCE_LOGS 		\
			$MAINTENANCE_WORK 		
do
	F_DIR_CHECK $dir $DEBUG_MODE || die $LINENO 4
done
F_READ_TEST ${MAINTENANCE_DIR}/generate_integrity_checks.sql $DEBUG_MODE || die $LINENO 5

# With these basic checks out of the way, direct all output to a timestamped log file.
LOG=${MAINTENANCE_LOGS}/${script}.${DBNAME}.${now}.log
F_INIT_OUTPUT_FILE $INSTNAME $LOG $DEBUG_MODE || die $LINENO 6
export LOG				 
F_REDIRECT_OUTPUT $LOG $DEBUG_MODE || die $LINENO 7

# Check no offline or online backup is running before connecting.
# This script could run concurrently against an online backup, but that would suggest a scheduling error. 
db2 -v attach to $INSTNAME || die $LINENO 8
while (( counter > 0 ))
do
    counter=$(db2 list utilities | grep -ic offline) || : #suppress trap of grep return code 1 if no match found
    (( counter > 0 )) && F_MESSAGE "Backup(s) are running." PLAIN I $DEBUG_MODE
    (( counter > 0 )) && sleep 300
done

export DB2NODE=0 # only meaningful in DPF
F_TEST_DBNAME  $DBNAME $DEBUG_MODE && F_CONNECT $DBNAME $DEBUG_MODE || die $LINENO 9

# Disable trap from here on, since we want explicit return code checks for DB2 commands and SQL queries. 
trap " " ERR

# Generate a file of SQL queries for each child application table in scope:
F_RUN_DB2FILE ${MAINTENANCE_DIR}/generate_integrity_checks.sql "-txf" 1 ${MAINTENANCE_WORK}/${script}_integrity_check.sql.${now} nocompress $DEBUG_MODE
RC=$?
case $RC in
     0) # run the generated script
		F_RUN_DB2FILE ${MAINTENANCE_WORK}/${script}_integrity_check.sql.${now}  "-tvf" 1 NONE  nocompress $DEBUG_MODE
		RC=$?; 
		(( RC > 1 )) && F_MESSAGE "CLP runtime error, please review this job log." PLAIN I $DEBUG_MODE && die $LINENO 10
		(( RC == 1 )) && F_MESSAGE "Referential Integrity checks pass at  database ${DBNAME} at timestamp ${now}." PLAIN I $DEBUG_MODE
		(( RC == 0 )) && F_MESSAGE "There is at least one implied Referential Integrity violation - please review ${LOG}." PLAIN I $DEBUG_MODE && die $LINENO 11
		;;
	 1)	F_MESSAGE "Nothing to do but this smells: there are no application-level declarative referential constraints at database ${DBNAME} at timestamp ${now}." BOLD I $DEBUG_MODE
		;;
	 *)	F_MESSAGE "SQL runtime error, please review this job log." PLAIN I $DEBUG_MODE && die $LINENO 12
		;;
esac

F_TERMINATE $DEBUG_MODE  

# Restore stdout and stderr to terminal
F_RESET_OUTPUT 
cat $LOG
exit $RC
