#!/bin/ksh93
#set tabstop=4 to display correctly in vi
#
# File Name:    PruneFiles.ksh (ksh93)
#               Based on existing copyleft code at website http://www.sustainablesoftware.net - please respect these terms.
# Author:       Alex Levy t/a http://www.sustainablesoftware.net
#
# Purpose:      1) Archive/Purge old files in the instance owner's home directory.
#               2) Archive and compress the db2diag, notify, archive and STMM logs.
#               3) Delete old archives
#               4) Archive and delete logs from SQL replication, if there are any.
#
# Out of scope: 5) Backup images on disk or tape
#               6) Recovery History File is pruned in a separate job, PruneHistory.ksh.
#
# Usage:        PruneFiles.ksh instance_name database_name
#               Assumes runner is the instance owner.
#               Set retention periods below - variables have name pattern *keepdays.
#
# File dependencies:
#               ${INSTHOME}/sqllib/db2profile
#               ${INSTHOME}/sqllib/userprofile
#               ${INSTHOME}/FUNCTIONS   (the userprofile exports $FPATH).
#               ${INSTHOME}/MAINTENANCE (the userprofile exports $MAINTENANCE_DIR).
#               ${INSTHOME}/MAINTENANCE/log  (the userprofile exports $MAINTENANCE_LOGS).
#               ${INSTHOME}/MAINTENANCE/log/work (the userprofile exports $MAINTENANCE_WORK).
#               DIAGPATH (as defined in the dbm cfg)
#               DIAGPATH/CFG (configuration capture directory, the userprofile exports $CFG_DIR)
#
# Calls:        ${MAINTENANCE}/Exit.ksh on error
#
# File dependencies for partitioned databases:
#               None - the working assumption is that $INSTHOME and the DIAGPATH are network-mounted
#                      and that no per-partition files are in scope for this job.
#
# Author: Alex Levy t/a http;//www.sustainablesoftware.net	
# Last Revision: June 2013.
#-------------------------------------------------------------------------------

# Functions in $FPATH never exit, they only return.
# We can either test each return code or use a trap. If we do use a trap, and we wish to test the return code first,
# then the test must be part of the same statement, e.g. F_CONNECT $DBNAME $DEBUG_MODE || die.
# The trap statement and the 2 inline functions 'FatalError' and 'die' handle the exit for all error conditions.
# The 'die' function is called on any error detected by the program logic.
# The FatalError function is triggered by any non-zero signal, including the exit code from 'die'.

trap FatalError ERR

function FatalError {
# Single point of exit on any error condition.
# This relies on a signal trap. ERR is any non-zero exit status and is Korn shell only.
# See 'man trap' for further detail.
# Can also call this function by 'kill -s ERR $$ 2>/dev/null'

RC=$?
if   (( RC == 99 ))
then # error has already been trapped by function 'die', do nothing more except exit
     ${MAINTENANCE_DIR}/Exit.ksh 99
else
     print ERR signal received
     if   [[ $LOG == "NOLOG" || ! -f $LOG || ! -s $LOG ]]
     then F_MESSAGE "$script Ends with error(s), see stdout above." PLAIN I +x
     else F_RESET_OUTPUT
          F_MESSAGE "$script ends with error(s). Please review log file $LOG . " PLAIN I +x | tee -a $LOG
          F_MESSAGE "The log file is displayed above for ease of reference. " PLAIN I +x | tee -a $LOG
     fi
     ${MAINTENANCE_DIR}/Exit.ksh 4
fi
}

function die {
# Single point of exit on any error condition caught by the program logic.
# Restore stdout and stderr if required, before flowing to FatalError.

LOG=${LOG:=NOLOG} # use $LOG if set, otherwise assign a value of NOLOG to $LOG.
LINE=${1:=unknown}
LOCATION=${2:=unknown}

if   [[ $LOG == "NOLOG" || ! -f $LOG || ! -s $LOG ]]
then F_MESSAGE "$script Ends with error(s) at line $LINE, location $LOCATION; see stdout above." PLAIN I +x
else F_RESET_OUTPUT
     cat $LOG
     F_MESSAGE "$script ends with error(s). Please review log file $LOG . " PLAIN I +x | tee -a $LOG
     F_MESSAGE "The log file is displayed above for ease of reference. " PLAIN I +x | tee -a $LOG
fi

# trigger FatalError
exit 99
}

#-------------------------------------------------------------------------------
# Declare and test variables.
#-------------------------------------------------------------------------------

set -u                              # assert all variables are defined and set.
typeset script=$(basename $0 .ksh)  # Strip off path and '.ksh' extension.
DEBUG_MODE="+x"                     # Default value - disable debugging.
typeset -i RC=0                 # return code from function calls and o/s
typeset -i sql_keepdays=7       # Edit! retention period for sqllib/tmp LOAD and IMPORT files
typeset -i maint_keepdays=14    # Edit! retention period for maintenance logs and workfiles
typeset -i diag_keepdays=14     # Edit! retention period for diagpath files and archives
typeset -i repl_keepdays=14     # Edit! retention period for replication log files
now=$(date +%Y%m%d%H%M%S)
typeset LOG=                    # set later
typeset DIAGPATH=               # set later

# Check the call
(( $# == 2 )) || F_ERROR_MESSAGE " Usage! I said:  $0 <instance_name> <database_name>" $DEBUG_MODE
(( $# == 2 )) || die $LINENO 1
typeset -l  INSTNAME=$1
typeset -u  DBNAME=$2
if   [[ $INSTNAME == $USER ]]
then :
else
     F_ERROR_MESSAGE "$0 : The runner must be the DB2 instance owner. Exiting ..." $DEBUG_MODE
     die $LINENO 2
fi

# Assert all required files and directories are in place and are accessible or executable;
# the first function call tests for $INSTHOME and the db2profile and executes db2profile.
F_SETUP_DB2_ENVIRONMENT $INSTHOME $INSTNAME $DBNAME $DEBUG_MODE || die  $LINENO 3
F_DIR_CHECK $MAINTENANCE_DIR  $DEBUG_MODE || die  $LINENO 6
F_DIR_CHECK $MAINTENANCE_LOGS $DEBUG_MODE || die  $LINENO 7
F_DIR_CHECK $MAINTENANCE_WORK $DEBUG_MODE || die  $LINENO 8
F_DIR_CHECK $CFG_DIR $DEBUG_MODE || die $LINENO 8_1

F_CONNECT $DBNAME $DEBUG_MODE || die $LINENO 9
# temporarily disable trap, since we want an explicit return code check
trap " " ERR
F_RUN_SQL "select value from  sysibmadm.dbmcfg where name = 'diagpath'" "-xnq" 0 $DEBUG_MODE
RC=$?; F_TEST_DB2_RETURNCODE $RC 0 $0 10 "Error querying sysibmadm.dbmcfg, DIAGPATH variables contains $DIAGPATH " || die $LINENO 10
DIAGPATH=${SQLVAR1%\/*} # remove any trailing '/' 
print DIAGPATH IS $DIAGPATH

trap FatalError ERR
F_DIR_CHECK ${DIAGPATH} $DEBUG_MODE || die  $LINENO 11
F_TERMINATE $DEBUG_MODE

# With these basic checks out of the way, direct all output to a timestamped log file.
LOG=${MAINTENANCE_LOGS}/${script}.${DBNAME}.${now}.log
F_INIT_OUTPUT_FILE $INSTNAME $LOG $DEBUG_MODE || die $LINENO 12
export LOG
F_REDIRECT_OUTPUT $LOG $DEBUG_MODE || die  $LINENO 13

# Capture filesystem utilisation before and after the prune
print "$INSTHOME and diagnostic log path utilisation before the prune:"
df -m $INSTHOME $DIAGPATH | sort -ur # sort in case the 2 paths are the same filesystem
F_TEST_RETURNCODE $? 0 $DEBUG_MODE || die  $LINENO 14

# $INSTHOME prunes
print "$script : 1. Pruning $INSTHOME/sqllib/tmp ..."
# these files are left over from calls to sysproc.admin_cmd; by default they log to sqllib/tmp
# but registry variable DB2_UTIL_MSGPATH can redirect them to another location.
find $INSTHOME/sqllib/tmp -name "IMPORT*" -type f -atime +${sql_keepdays} -exec rm -f {} \; > /dev/null
find $INSTHOME/sqllib/tmp -name "LOAD*"   -type f -atime +${sql_keepdays} -exec rm -f {} \; > /dev/null

if   (( SQL_REPLICATION_SOURCE == 0 && SQL_REPLICATION_TARGET == 0 )) # the userpofile exports these variables
then print "$script : 2. No SQL replication logs to prune."
else
     print "$script : 2. Archiving SQL replication logs ... "
     # db2inst1  2966  2965  0 Apr04 ?        00:07:29 asncap capture_server=SAMPLE capture_schema=ASN capture_path=/home/db2inst1/replication/capture_files startmode=WARMSI
     # db2inst2 28002     1  1 Apr04 ?        02:47:25 /home/db2inst2/sqllib/bin/asnapply apply_qual=AQ00 control_server=samplrep apply_path=/database/replication pwdfile=asnpwd.aut
     typeset repl_logpath=$(ps -eaf | egrep "asncap|asnapply" | grep -v grep | awk -F"=" '{print $4}' | awk '{print $1}')
     F_DIR_CHECK ${repl_logpath} $DEBUG_MODE || die $LINENO 14_1
     # SQL Replication does not rotate log files or respawn; we therefore need to archive a copy of the log file, leaving the original intact.
     for file in $(ls -1  ${repl_logpath}/*.log)
     do
         cp -p $file ${CFG_DIR}/${DBNAME}.replication_log.${now}
         F_ZIP_IT ${CFG_DIR}/${DBNAME}.replication_log.${now} $DEBUG_MODE || die $LINENO 14_2
     done
fi

print "$script : 3. Maintenance directory prunes"
find $MAINTENANCE_LOGS -type f -mtime +${maint_keepdays} -exec rm -f {} \;  > /dev/null
find $MAINTENANCE_WORK -type f -mtime +${maint_keepdays} -exec rm -f {} \;  > /dev/null

print "$script : 4. Pruning diagnostic path $DIAGPATH "
[[ -r ${DIAGPATH}/db2diag.log ]] && db2diag -A ${DIAGPATH}/db2diag.log
RC=$? ; F_TEST_RETURNCODE $RC 1 $DEBUG_MODE || die $LINENO 16
[[ -r ${DIAGPATH}/$INSTNAME.nfy ]] && db2diag -A ${DIAGPATH}/$INSTNAME.nfy
RC=$? ; F_TEST_RETURNCODE $RC 1 $DEBUG_MODE || die $LINENO 17
# Allow for no event or STMM logs (rc=2 from db2diag)
db2diag -A ${DIAGPATH}/events/*.log
RC=$? ; F_TEST_RETURNCODE $RC 2 $DEBUG_MODE || die $LINENO 18
db2diag -A ${DIAGPATH}/stmmlog/stmm.*.log
RC=$? ; F_TEST_RETURNCODE $RC 2 $DEBUG_MODE || die $LINENO 19

for file in $(ls -1 ${DIAGPATH}/db2diag.log_*  ${DIAGPATH}/${INSTNAME}.nfy_* ${DIAGPATH}/events/*.log_* ${DIAGPATH}/stmmlog/stmm.*.log_* | grep -v gz)
do
    F_ZIP_IT ${file} $DEBUG_MODE  || die $LINENO 20
done

print "$script : 5. Archive db2diag.log, $INSTNAME.nfy, event and STMM logs."
mkdir -m 755 ${DIAGPATH}/ARCHIVE.${now}
mv ${DIAGPATH}/*.gz ${DIAGPATH}/ARCHIVE.${now}
mv ${DIAGPATH}/events/*.gz  ${DIAGPATH}/ARCHIVE.${now}
mv ${DIAGPATH}/stmmlog/*.gz ${DIAGPATH}/ARCHIVE.${now}

# Remove any flat files and archives older than $diag_keepdays days old
find ${DIAGPATH} -type f -mtime +${diag_keepdays} -exec rm -f {} \;
RC=$? ; F_TEST_RETURNCODE $? 1 $DEBUG_MODE || die $LINENO 21
find ${DIAGPATH}/CFG -type f -mtime +${diag_keepdays} -exec rm -f {} \;
RC=$? ; F_TEST_RETURNCODE $? 1 $DEBUG_MODE || die $LINENO 22
find ${DIAGPATH} -type d -name "ARCHIVE.*" -mtime +${diag_keepdays} -exec rm -fr {} \;
RC=$? ; F_TEST_RETURNCODE $? 1 $DEBUG_MODE || die $LINENO 221
# and core dump directories - no need to retain these assuming db2support has been run
find ${DIAGPATH}/ -name "c*" -type d -exec rm -fr {} \;

print "$INSTHOME and diagnostic log path utilisation after the prune:"
trap " " ERR
df -m $INSTHOME $DIAGPATH | sort -ur # sort in case the 2 paths are the same filesystem
RC=$? ; F_TEST_RETURNCODE $? 0 $DEBUG_MODE || die $LINENO 23

# Restore stdout and stderr to terminal
F_RESET_OUTPUT
cat $LOG
exit 0

