#!/bin/ksh93
#set tabstop=4 to display correctly in vi
#
#-------------------------------------------------------------------------------
# File Name:    PruneHistory.ksh (ksh93)
#				Based on existing copyleft code at website http://www.sustainablesoftware.net - please respect these terms.
# Author:       Alex Levy t/a http://www.sustainablesoftware.net
#
# Contributors: Steve Harris (Thanks Steve!)
#
# Purpose:      Synchronise history prunes with backups, i.e. prune history in step with deleted backup images.
#               DB2 will automatically prune the history at the end of a successful backup if the number of backups
#               in the history >= NUM_DB_BACKUPS *but* only for history older than REC_HIS_RETENTN days. 
#               However the RFH is per partition therefore the number of 'successful' backups is not necessarily consistent.
#               Further REC_HIS_RETENTN is not necessarily consistent with NUM_DB_BACKUPS since backups may fail, or may not be regularly scheduled.
#               We also need to ignore 'successful' backups to /dev/null, e.g. when transitioning a database to archive logging.
#               Finally the RFH is per partition therefore the number of 'successful' backups is not necessarily consistent.
#               Allow for the recovery history file changing during a list history backup (SQL2155W).
# 
# Out of scope: 5) Backup images on disk or tape
#               6) Recovery History File is pruned in a separate job, PruneHistory.ksh.
# 
# Usage:        PruneHistory.ksh instance_name database_name 
#               Assumes runner is the instance owner.
#
# File dependencies:
#               ${INSTHOME}/sqllib/db2profile
#               ${INSTHOME}/sqllib/userprofile 
#               ${INSTHOME}/sqllib/db2nodes.cfg 
#               ${INSTHOME}/FUNCTIONS   (the userprofile exports $FPATH).
#				${INSTHOME}/MAINTENANCE (the userprofile exports $MAINTENANCE_DIR).
#				${INSTHOME}/MAINTENANCE/log  (the userprofile exports $MAINTENANCE_LOGS).
#				${INSTHOME}/MAINTENANCE/log/work (the userprofile exports $MAINTENANCE_WORK).
#
# Calls:		${MAINTENANCE}/Exit.ksh on error	
#
# File dependencies for partitioned databases:
#               None - the working assumption is that $INSTHOME is network-mounted on DPF systems.
#
# Notes:        Customised for the BBC: 
#               There is no need to delete archive logs in step, because:
#               a) The Commvault solution uses a staging library /opt/simpana/db2/logfiles/${dbalias}/${node} and can delete archive logs on disk automatically.
#               b) All pre-stage environments and all reporting servers use circular logging only.
#				Future-proof for any DPF instances, e.g. a future reporting data warehouse.
#
# Date:         April 2012.           
#-------------------------------------------------------------------------------

# Functions in $FPATH never exit, they only return.
# We can either test each return code or use a trap. If we do use a trap, and we wish to test the return code first,
# then the test must be part of the same statement, e.g. F_CONNECT $DBNAME $DEBUG_MODE || die.
# The trap statement and the 2 inline functions 'FatalError' and 'die' handle the exit for all error conditions.
# The 'die' function is called on any error detected by the program logic.
# The FatalError function is triggered by any non-zero signal, including the exit code from 'die'.

trap FatalError ERR

function FatalError {
# Single point of exit on any error condition.
# This relies on a signal trap. ERR is any non-zero exit status and is Korn shell only. 
# See 'man trap' for further detail.
# Can also call this function by 'kill -s ERR $$ 2>/dev/null'

RC=$? 
if   (( RC == 99 ))
then # error has already been trapped by function 'die', do nothing more except exit
	 ${MAINTENANCE_DIR}/Exit.ksh 99	
else
     print ERR signal received
	 if   [[ $LOG == "NOLOG" || ! -f $LOG || ! -s $LOG ]]
	 then F_MESSAGE "$script Ends with error(s), see stdout above." PLAIN I +x
	 else F_RESET_OUTPUT
     	  cat $LOG
     	  F_MESSAGE "$script ends with error(s). Please review log file $LOG . " PLAIN I +x | tee -a $LOG
     	  F_MESSAGE "The log file is displayed above for ease of reference. " PLAIN I +x | tee -a $LOG
	 fi
	 ${MAINTENANCE_DIR}/Exit.ksh 4	
fi
}

function die {
# Single point of exit on any error condition caught by the program logic.
# Restore stdout and stderr if required, before flowing to FatalError.

LOG=${LOG:=NOLOG} # use $LOG if set, otherwise assign a value of NOLOG to $LOG.
LINE=${1:=unknown}
LOCATION=${2:=unknown}

if   [[ $LOG == "NOLOG" || ! -f $LOG || ! -s $LOG ]]
then F_MESSAGE "$script Ends with error(s) at line $LINE, location $LOCATION; see stdout above." PLAIN I +x
else F_RESET_OUTPUT
     cat $LOG
     F_MESSAGE "$script ends with error(s). Please review log file $LOG . " PLAIN I +x | tee -a $LOG
     F_MESSAGE "The log file is displayed above for ease of reference. " PLAIN I +x | tee -a $LOG
fi

# trigger FatalError
exit 99
}

#-------------------------------------------------------------------------------
# Declare and test variables.
#-------------------------------------------------------------------------------

set -u								# assert all variables are defined and set.	 
typeset script=$(basename $0 .ksh)  # Strip off path and '.ksh' extension.
DEBUG_MODE="+x"                     # Default value - disable debugging.
typeset -i RC=0						# return code from function calls and o/s
typeset -i num_backups_to_keep=-1 
typeset -i num_partitions=-1
typeset -i count=99 
typeset -i no_loops=0			# loop counter for attempts to read the RHF.
now=$(date +%Y%m%d%H%M%S)              
typeset timestmp=               # set later, holds timestamp of the oldest backup we retain.
typeset LOG=NOLOG               # set later

# Check the call
(( $# == 2 )) || F_ERROR_MESSAGE " Usage! I said:  $0 <instance_name> <database_name>" $DEBUG_MODE   
(( $# == 2 )) || die $LINENO 1 
typeset -l  INSTNAME=$1
typeset -u  DBNAME=$2
if   [[ $INSTNAME == $USER ]]
then :
else
     F_ERROR_MESSAGE "$0 : The runner must be the DB2 instance owner. Exiting ..." $DEBUG_MODE
     die $LINENO 2
fi

# Assert all required files and directories are in place and are accessible or executable;
# the first function call tests for $INSTHOME and the db2profile and executes db2profile.
F_SETUP_DB2_ENVIRONMENT $INSTHOME $INSTNAME $DBNAME $DEBUG_MODE || die  $LINENO 3
F_DIR_CHECK ${INSTHOME}/deploy_db/bin $DEBUG_MODE || die  $LINENO 4
F_DIR_CHECK $MAINTENANCE_DIR  $DEBUG_MODE   || die $LINENO 6
F_DIR_CHECK $MAINTENANCE_LOGS $DEBUG_MODE   || die $LINENO 7
F_DIR_CHECK $MAINTENANCE_WORK $DEBUG_MODE   || die $LINENO 8 
F_READ_TEST ${INSTHOME}/sqllib/db2nodes.cfg $DEBUG_MODE || die $LINENO 9

# With these basic checks out of the way, direct all output to a timestamped log file.
LOG=${MAINTENANCE_LOGS}/${script}.${DBNAME}.${now}.log
F_INIT_OUTPUT_FILE $INSTNAME $LOG $DEBUG_MODE || die $LINENO 10
export LOG				 
F_REDIRECT_OUTPUT $LOG $DEBUG_MODE || die $LINENO 11

# check no backup is running before connecting
db2 -v attach to $INSTNAME || die $LINENO 10
while (( count > 0 ))
do
    count=$(db2 list utilities | grep -ic backup) || : #suppress Unix trap of return code 1 if no match found
    (( count > 0 )) && F_MESSAGE "Backup(s) are running." PLAIN I $DEBUG_MODE
    (( count > 0 )) && sleep 300
done

export DB2NODE=0 # only meaningful in DPF
F_CONNECT $DBNAME $DEBUG_MODE || die $LINENO 12
 
# disable trap from here on, since we want explicit return code checks for SQL queries 
# and also want to tolerate errors from db2 list history.
trap " " ERR

# Check every partition has the same NUM_DB_BACKUPS 

F_RUN_SQL "select numback from (select dbpartitionnum, smallint(value)as numback from sysibmadm.dbcfg where name = 'num_db_backups') zzzz \
		   group by numback having min(numback) = max(numback)" \
                 "-xnq" 0 $DEBUG_MODE
RC=$? ; F_TEST_DB2_RETURNCODE $RC 0 $0 13 "DB CFG error, num_db_backup is not consistent across partitions." || die $LINENO 13
num_backups_to_keep=${SQLVAR1}

# Get the number of partitions for this instance
F_RUN_SQL "select num_dbpartitions from sysibmadm.env_inst_info where inst_name = '${INSTNAME}'" "-xnq" 0 $DEBUG_MODE
RC=$? ; F_TEST_DB2_RETURNCODE $RC 0 $0 14 "Query error against sysibmadm.env_inst_info." || die $LINENO 14
num_parts=${SQLVAR1}

# Report history volumes before the prune
F_RUN_QUERY "select 'There were    ',count(*),' history events between ',min(start_time),' and ',max(start_time),'.' from sysibmadm.db_history" "-x" 0 $DEBUG_MODE 
RC=$? ; F_TEST_DB2_RETURNCODE $RC 0 $0 15 "Query error against sysibmadm.db_history." || die $LINENO 15

# Prune the history file, if and only if the backup corresponding to NUM_DB_BACKUPS succeeded on EVERY partition.
# Do not count backups to /dev/null 
if (( num_backups_to_keep > 0 ))
then
   # remove all history from 1 second before the N th successful backup
   # or some default date/time in the past if we have less than n backups.
   # The sub select only returns the row for the N th backup. 
   # If less than N backups exist, then it returns no rows,
   # in which case we need an aggregate function such as max to expose and coalesce the null 
   # into a default date/time of '19171025120000' (approximate timestamp of Bolshevik Revolution).
   # operation = 'B' = backup
   # objecttype = 'D' = full backup
   # sqlcode is null  -- bit odd this, it's only populated on failure
   # sqlwarn is null -- ditto
   F_RUN_SQL         "select coalesce(max(start), 19171025120000) from 			\
                        (select bigint(start_time) - 1 as start, 				\
                        row_number() over(order by start_time desc) as backup 	\
                        from   sysibmadm.db_history              				\
                        where  operation = 'B'                   				\
                        and    objecttype = 'D'                  				\
                        and    sqlcode is null                   				\
                        and    sqlwarn is null                   				\
						and    location <> '/dev/null'							\
						group by start_time having count(*) = $num_parts 		\
                        ) as zzz                                 				\
                        where backup = $num_backups_to_keep"   "-x" 0 $DEBUG_MODE
   RC=$? ; F_TEST_DB2_RETURNCODE $RC 0 $0 16 "Query error against sysibmadm.db_history." || die $LINENO 16
   timestmp=${SQLVAR1}
   if   (( num_parts == 1 ))
   then F_RUN_DB2CLP "prune history $timestmp and delete" 0 NONE nocompress noattach $DEBUG_MODE	
        RC=$? ; F_TEST_DB2_RETURNCODE $RC 0 $0 17 "Prune error." || die $LINENO 17
   else # I haven't yet tested db2_all against the latest $FPATH function library
        db2_all "\"|| db2 connect to $DBNAME; db2 -v prune history $timestmp and delete; db2 terminate"
        RC=$? ; F_TEST_DB2_RETURNCODE $RC 0 $0 18 "Prune error." || die $LINENO 18
   fi

   # loop until the recovery history file is stable and then report it 
   RC=999
   while (( RC > 0 ))
   do 
      if   (( num_parts == 1 )) 
      then 
           F_RUN_DB2CLP "list history backup since $timestmp for $DBNAME" 0 ${MAINTENANCE_LOGS}/backuphistory.${now}  nocompress noattach $DEBUG_MODE > /dev/null
      else 
           db2_all "<<+0< db2 -v list history backup since $timestmp for $DBNAME" > ${MAINTENANCE_LOGS}/backuphistory.${now} > /dev/null
      fi
      RC=$?
      print RC for list history was $RC
      if (( no_loops > 720 ))
      then
         # then you've been waiting an hour 
         print $0 "I'm tired of waiting for the recovery history file to stabilise. I'm giving up"
         break
      else
         sleep 5
         let no_loops=no_loops+1
      fi
   done
   echo "Content of backhist file:"
   cat ${MAINTENANCE_LOGS}/backuphistory.${now}
fi

# Report history volumes after the prune
F_RUN_QUERY "select 'There are now ',count(*),' history events between ',min(start_time),' and ',max(start_time),'.' from sysibmadm.db_history" "-x" 0 $DEBUG_MODE 
RC=$? ; F_TEST_DB2_RETURNCODE $RC 0 $0 19 "Query error against sysibmadm.db_history." || die $LINENO 19
F_TERMINATE $DEBUG_MODE  

# Restore stdout and stderr to terminal
F_RESET_OUTPUT 
cat $LOG
exit 0
