#!/bin/ksh93
#set tabstop=4 to display correctly in vi
#
#-------------------------------------------------------------------------------
# File Name:    Reorg_Online.ksh (ksh93)
#               Based on existing copyleft code at website http://www.sustainablesoftware.net - please respect these terms.
#
# Purpose:      Selective REORG of tables and indexes online
#               a. if flagged by REORGCHK and
#               b. if the table or index schema is in the user-supplied list.
#
# Usage:        Reorg_Online.ksh $1 instance_name $2 database_name
#               Assumes runner is the instance owner.
#
# Out of scope:
#
# File dependencies:
#               ${INSTHOME}/sqllib/db2profile
#               ${INSTHOME}/sqllib/userprofile
#               ${INSTHOME}/FUNCTIONS            - the userprofile exports this as $FPATH.
#               ${INSTHOME}/MAINTENANCE          - the userprofile exports this as $MAINTENANCE_DIR.
#               ${INSTHOME}/MAINTENANCE/log      - the userprofile exports this as $MAINTENANCE_LOGS.
#               ${INSTHOME}/MAINTENANCE/log/work - the userprofile exports this as $MAINTENANCE_WORK.
#               ${INSTHOME}/MAINTENANCE/REORGS   - directory for this script.
#               ${INSTHOME}/MAINTENANCE/REORGS/reorgchk_drop.ddl                    - DDL for work tables
#               ${INSTHOME}/MAINTENANCE/REORGS/reorgchk_tb.ddl                      - "
#               ${INSTHOME}/MAINTENANCE/REORGS/reorgchk_ix.ddl                      - "
#               ${INSTHOME}/MAINTENANCE/REORGS/reorg.pick_table.sql                 - Queries against REORGCHK output
#               ${INSTHOME}/MAINTENANCE/REORGS/reorg.pick_index.sql                 - "
#               ${INSTHOME}/MAINTENANCE/REORGS/reorg.schema_list.<lowercase_dbname> - driver file of schema names, one per line.
#               ${MAINTENANCE_DIR}/TEMPLATES/reorgchk_tb.ddl.template               - DDL template
#               ${MAINTENANCE_DIR}/TEMPLATES/reorgchk_ix.ddl.template               - "
#               ${MAINTENANCE_DIR}/TEMPLATES/reorg_pick_table.sqltemplate           - SQL template
#               ${MAINTENANCE_DIR}/TEMPLATES/reorg_pick_index.sql.template          - SQL template
#
# File dependencies for partitioned databases:
#               None - REORG operates on all partitions.
#
# Calls:        ${MAINTENANCE}/Exit.ksh on error
#
# Notes:        Read A_SETUP_README in this directory for local site customisations.
#               In particular, decide read/write access whilst a REORG is in progress and check your comfort with INPLACE reorgs.
#               Also: Prune $MAINTENANCE_LOGS and $MAINTENANCE_WORK manually if you are not scheduling or running the PruneFiles.ksh script
#                     Comment out the REORGCHK UPDATE STATISTICS against each schema in reorg.schema.list.${DBNAME}, if you are satisfied catalog statistics are current.
#
# Date:          05.08.2005 and enhanced for publication on the web, 20.03.2006
# Modified       09.10.2007 for DB2 V9.1
#                04.05.2011 Modify reorg.pick_xxx DDL scripts to run REORGs with INPLACE and ALLOW WRITE ACCESS -
#                           unless the tables are range partitioned or MDCs.
#                           SNAPSHOT_REORG is deprecated and replaced by SNAPTAB_REORG. Will add to the reorg.pick_xxx DDL scripts.
#                           Use runner's implicit schema rather than dbadmin for work tables; modify RUNSTATS of control tables in line
#                           Change all 'echo' to 'print'
#                02.05.2012 Re-integrate into a complete maintenance suite for DB2 9.7;
#                           add extra XML and data partition support (the latter saw several enhancvements at V9.7);
#                           recast script TO ksh93.
#				 04.03.2013 Add schema name to output files; allow for unindexed schemata (though this breaks sound model design).
#
# Author:       Alex Levy t/a http://www.sustainablesoftware.net
# Credits:      Dan 'The Man' Day
# Date:         March 2013.
#-------------------------------------------------------------------------------

# Functions in $FPATH never exit, they only return.
# We can either test each return code or use a trap. If we do use a trap, and we wish to test the return code first,
# then the test must be part of the same statement, e.g. F_CONNECT $DBNAME $DEBUG_MODE || die.
# The trap statement and the 2 inline functions 'FatalError' and 'die' can handle the exit for all error conditions.
# The 'die' function is called on any error detected by the program logic.
# The FatalError function is triggered by any non-zero signal, including the exit code from 'die'.

trap FatalError ERR

function FatalError {
# Single point of exit on any error condition.
# This relies on a signal trap. ERR is any non-zero exit status and is Korn shell only.
# See 'man trap' for further detail.
# Can also call this function by 'kill -s ERR $$ 2>/dev/null'

RC=$?
if   (( RC == 99 ))
then # error has already been trapped by function 'die', do nothing more except exit
     ${MAINTENANCE_DIR}/Exit.ksh 99
else
     print ERR signal received
     if   [[ $LOG == "NOLOG" || ! -f $LOG || ! -s $LOG ]]
     then F_MESSAGE "$script Ends with error(s), see stdout above." PLAIN I +x
     else F_RESET_OUTPUT
          cat $LOG
          F_MESSAGE "$script ends with error(s). Please review log file $LOG . " PLAIN I +x | tee -a $LOG
          F_MESSAGE "The log file is displayed above for ease of reference. " PLAIN I +x | tee -a $LOG
     fi
     ${MAINTENANCE_DIR}/Exit.ksh 4
fi
}

function die {
# Single point of exit on any error condition caught by the program logic.
# Restore stdout and stderr if required, before flowing to FatalError.

LOG=${LOG:=NOLOG} # use $LOG if set, otherwise assign a value of NOLOG to $LOG.
LINE=${1:=unknown}
LOCATION=${2:=unknown}

if   [[ $LOG == "NOLOG" || ! -f $LOG || ! -s $LOG ]]
then F_MESSAGE "$script Ends with error(s) at line $LINE, location $LOCATION; see stdout above." PLAIN I +x
else F_RESET_OUTPUT
     cat $LOG
     F_MESSAGE "$script ends with error(s) at line $LINE, location ${LOCATION}. Please review log file $LOG . " PLAIN I +x | tee -a $LOG
     F_MESSAGE "The log file is displayed above for ease of reference. " PLAIN I +x | tee -a $LOG
fi

# trigger FatalError
exit 99
}

#-------------------------------------------------------------------------------
# Declare and test variables.
#-------------------------------------------------------------------------------

set -u                              # assert all variables are defined and set.
typeset script=$(basename $0 .ksh)  # Strip off path and '.ksh' extension.
DEBUG_MODE="+x"                     # Default value - disable debugging.
typeset -i RC=0                     # return code from function calls and o/s
typeset -i counter=999
now=$(date +%Y%m%d%H%M%S)
typeset LOG=NOLOG                   # set later
typeset -l SCHEMA=                  # "

# Check the call
(( $# == 2 )) || F_ERROR_MESSAGE " Usage! I said:  $0 <instance_name> <database_name>" $DEBUG_MODE
(( $# == 2 )) || die $LINENO 1
typeset -l  INSTNAME=$1
typeset -l  DBNAME=$2
if   [[ $INSTNAME == $USER ]]
then :
else
     F_ERROR_MESSAGE "$0 : The runner must be the DB2 instance owner. Exiting ..." $DEBUG_MODE
     die $LINENO 2
fi

# Assert all required files and directories are in place and are accessible or executable;
# the first function call tests for $INSTHOME and the db2profile and executes db2profile.
F_SETUP_DB2_ENVIRONMENT $INSTHOME $INSTNAME $DBNAME $DEBUG_MODE             || die $LINENO 3
for dir in  $FPATH                  \
            $MAINTENANCE_DIR        \
            $MAINTENANCE_LOGS       \
            $MAINTENANCE_WORK       \
            $MAINTENANCE_DIR/REORGS \
            $MAINTENANCE_DIR/TEMPLATES
do
    F_DIR_CHECK $dir $DEBUG_MODE || die $LINENO 4
done
for file in ${MAINTENANCE_DIR}/REORGS/reorg.schema_list.$DBNAME         \
            ${MAINTENANCE_DIR}/REORGS/reorgchk_drop.ddl                 \
            ${MAINTENANCE_DIR}/REORGS/reorgchk_tb.ddl                   \
            ${MAINTENANCE_DIR}/REORGS/reorgchk_ix.ddl                   \
            ${MAINTENANCE_DIR}/REORGS/reorg.pick_table.sql              \
            ${MAINTENANCE_DIR}/REORGS/reorg.pick_index.sql              \
            ${MAINTENANCE_DIR}/TEMPLATES/reorgchk_tb.ddl.template       \
            ${MAINTENANCE_DIR}/TEMPLATES/reorgchk_ix.ddl.template       \
            ${MAINTENANCE_DIR}/TEMPLATES/reorg.pick_table.sql.template  \
            ${MAINTENANCE_DIR}/TEMPLATES/reorg.pick_index.sql.template
do
    F_READ_TEST $file $DEBUG_MODE || die $LINENO 5
done

# With these basic checks out of the way, direct all output to a timestamped log file.
LOG=${MAINTENANCE_LOGS}/${script}.${DBNAME}.${now}.log
F_INIT_OUTPUT_FILE $INSTNAME $LOG $DEBUG_MODE || die $LINENO 6
export LOG
F_REDIRECT_OUTPUT $LOG $DEBUG_MODE || die $LINENO 7

# Check no offline or online backup is running before connecting.
# This script could run concurrently against an online backup, but that would suggest a scheduling error.
db2 -v attach to $INSTNAME || die $LINENO 8
while (( counter > 0 ))
do
    counter=$(db2 list utilities | grep -ic offline) || : #suppress trap of grep return code 1 if no match found
    (( counter > 0 )) && F_MESSAGE "Backup(s) are running." PLAIN I $DEBUG_MODE
    (( counter > 0 )) && sleep 300
done

export DB2NODE=0 # only meaningful in DPF
F_CONNECT $DBNAME $DEBUG_MODE || die $LINENO 9

# Disable trap from here on, since we want explicit return code checks for DB2 commands and SQL queries.
trap " " ERR

for SCHEMA in $(cat ${MAINTENANCE_DIR}/REORGS/reorg.schema_list.$DBNAME)
do
    typeset -u SCHEMA=$SCHEMA
    print "$0 : Running for schema $SCHEMA"
    # Test the schema exists and has at least one table.
    F_RUN_SQL "select 1 from sysibm.sysdummy1 where '${SCHEMA}' = any(select tabschema from syscat.tables where type = 'T')" "-xnq" 1 $DEBUG_MODE > /dev/null
    RC=$?
    case $RC in
         0) # comment out the reorgchk update statistics if you are satisfied statistics are up to date
            #F_RUN_DB2CLP "reorgchk update statistics on schema $SCHEMA" 0 ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_1stats.${now} nocompress noattach $DEBUG_MODE
            #RC=$? ; (( RC > 0 )) && F_MESSAGE "CLP Error, please review ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_1stats.${now}" PLAIN I $DEBUG_MODE && die $LINENO 10
            # Drop work tables from the previous run, if any.
            F_RUN_DB2FILE ${MAINTENANCE_DIR}/REORGS/reorgchk_drop.ddl "-tvf" 0  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_2drop_old.${now} nocompress $DEBUG_MODE
            RC=$?; (( RC > 0 )) && F_MESSAGE "SQL runtime error, please review  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_2drop_old.${now}" PLAIN I $DEBUG_MODE && die $LINENO 11
            # Recreate work tables for REORGCHK table stats
            F_RUN_DB2FILE ${MAINTENANCE_DIR}/REORGS/reorgchk_tb.ddl "+c -stvf" 0  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_3create_new.${now} nocompress $DEBUG_MODE
            RC=$?; (( RC > 0 )) && F_MESSAGE "SQL runtime error, please review ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_3create_new.${now}" PLAIN I $DEBUG_MODE && die $LINENO 12
            db2 -v commit
            # Get table REORGCHK stats; run with autocommit off as the procedure creates a DGTT, used as a template for the following create table statement.
            F_RUN_DB2CLP "+c -v CALL sysproc.reorgchk_tb_stats('S','${SCHEMA}')" 0 ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_4reorgchk_table.${now} nocompress noattach $DEBUG_MODE
            RC=$?; (( RC > 0 )) && F_MESSAGE "SQL runtime error, please review  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_4reorgchk_table.${now}" PLAIN I $DEBUG_MODE && die $LINENO 13
            # Populate work table with table REORGCHK stats
            print "insert into reorgchk_tb select * from session.tb_stats;" >  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_5populate.${now}
            F_RUN_DB2FILE ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_5populate.${now} "+c -tvf" 0 ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_5populate.out.${now} nocompress $DEBUG_MODE
            RC=$?; (( RC > 0 )) && F_MESSAGE "SQL runtime error, please review  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_5populate.out.${now}" PLAIN I $DEBUG_MODE && die $LINENO 14
            # Select tables in need of a REORG (formulae 1-3), picking the most appropriate index, and generate a reorg script.
            F_RUN_DB2FILE ${MAINTENANCE_DIR}/REORGS/reorg.pick_table.sql "+c -xtf" 1  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_6reorg.sql.${now} nocompress $DEBUG_MODE
            RC=$?; (( RC > 1 )) && F_MESSAGE "SQL runtime error, please review ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_6reorg.sql.${now}" PLAIN I $DEBUG_MODE && die $LINENO 15
            if  (( RC == 1 ))
            then F_MESSAGE "No tables in schema $SCHEMA require a REORG today." PLAIN I $DEBUG_MODE
                 RC=0
            else F_MESSAGE "Running table REORGs for schema $SCHEMA ..." PLAIN I $DEBUG_MODE
                 # Run the script file if it's populated - this script contains RUNSTATS and COMMIT statements as well as the REORGs themselves.
                 # First remove DB20000I messages from the generated file
                 cat  ${MAINTENANCE_WORK}/${script}_${DBNAME}_6reorg.sql.${now} | grep -v DB20000I >  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_6reorg.sql.${now}.edit
                 mv ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_6reorg.sql.${now}.edit ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_6reorg.sql.${now}
                 # Allow for SQL2314W, return code 2
                 F_RUN_DB2FILE ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_6reorg.sql.${now} "-stvf" 2 ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_7reorg.out.${now}  nocompress $DEBUG_MODE
                 RC=$?; (( RC > 2 )) && F_MESSAGE "CLP runtime error, please review  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_7reorg.out.${now}" PLAIN I $DEBUG_MODE && die $LINENO 16
                 F_MESSAGE "Table REORGs for schema $SCHEMA are complete." PLAIN I $DEBUG_MODE
            fi
            ###
            # Now for the index level REORGs ...these follow much the same procedural logic as table level REORGs.
            ###
            # It *should* be assumed, without further testing, that the schema has at least 1 index (otherwise the data model is at fault!). However:
    		F_RUN_SQL "select 1 from sysibm.sysdummy1 where '${SCHEMA}' = any(select indschema from syscat.indexes)" "-xnq" 1 $DEBUG_MODE > /dev/null
    		RC=$?
			if (( RC == 0 ))
			then
				# Recreate work tables for REORGCHK index stats
            	F_RUN_DB2FILE ${MAINTENANCE_DIR}/REORGS/reorgchk_ix.ddl "+c -stvf" 0  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_8create_new.${now} nocompress $DEBUG_MODE
            	RC=$?; (( RC > 0 )) && F_MESSAGE "SQL runtime error, please review ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_8create_new.${now}" PLAIN I $DEBUG_MODE && die $LINENO 17
            	# Get index REORGCHK stats; run with autocommit off as the procedure creates a DGTT, used as a template for the following create table statement.
            	F_RUN_DB2CLP "+c -v CALL sysproc.reorgchk_ix_stats('S','${SCHEMA}')" 0 ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_9reorgchk_index.${now} nocompress noattach $DEBUG_MODE
            	RC=$?; (( RC > 0 )) && F_MESSAGE "SQL runtime error, please review  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_9reorgchk_index.${now}" PLAIN I $DEBUG_MODE && die $LINENO 18
            	# Populate work table with index REORGCHK stats
            	print "insert into reorgchk_ix select * from session.ix_stats;" >  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_10populate.${now}
            	F_RUN_DB2FILE ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_10populate.${now} "+c -tvf" 0 ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_10populate.out.${now} nocompress $DEBUG_MODE
            	RC=$?; (( RC > 0 )) && F_MESSAGE "SQL runtime error, please review  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_10populate.out.${now}" PLAIN I $DEBUG_MODE && die $LINENO 19
            	# Generate a reorg script, based on formulae F4 - F8.
            	F_RUN_DB2FILE ${MAINTENANCE_DIR}/REORGS/reorg.pick_index.sql "+c -xtf" 1  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_11reorg.sql.${now} nocompress $DEBUG_MODE
            	RC=$?; (( RC > 1 )) && F_MESSAGE "SQL runtime error, please review ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_11reorg.sql.${now}" PLAIN I $DEBUG_MODE && die $LINENO 20
            	if  (( RC == 1 ))
            	then F_MESSAGE "No indices in schema $SCHEMA require a REORG today." PLAIN I $DEBUG_MODE
                 	RC=0
            	else F_MESSAGE "Running index REORGs for schema $SCHEMA ..." PLAIN I $DEBUG_MODE
                 	# Run the script file if it's populated - this script contains RUNSTATS and COMMIT statements as well as the REORGs themselves.
                 	# First remove DB20000I messages from the generated file
                 	cat  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_11reorg.sql.${now} | grep -v DB20000I >  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_11reorg.sql.${now}.edit
                 	mv ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_11reorg.sql.${now}.edit ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_11reorg.sql.${now}
                 	# Allow for SQL2314W, return code 2
                 	F_RUN_DB2FILE ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_11reorg.sql.${now} "-stvf" 2 ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_12reorg.out.${now}  nocompress $DEBUG_MODE
                 	RC=$?; (( RC > 2 )) && F_MESSAGE "CLP runtime error, please review  ${MAINTENANCE_WORK}/${script}_${DBNAME}_${SCHEMA}_12reorg.out.${now}" PLAIN I $DEBUG_MODE && die $LINENO 21
                 	F_MESSAGE "Index REORGs for schema $SCHEMA are complete." PLAIN I $DEBUG_MODE
            	fi
			else
        		F_MESSAGE "Table schema ${SCHEMA} has no indexes in the same schema at database ${DBNAME}. Please confirm." PLAIN I $DEBUG_MODE
				RC=0
			fi
            ;;
        1)  F_MESSAGE "Table schema ${SCHEMA} has no tables on database ${DBNAME}. Check parameter file ${MAINTENANCE_DIR}/REORGS/reorg.schema_list.$DBNAME for errors." PLAIN I $DEBUG_MODE
            ;;
        *)  F_MESSAGE "SQL runtime error, please review ${MAINTENANCE_WORK}/F_RUN_SQL.out" PLAIN I $DEBUG_MODE && die $LINENO 22
            ;;
     esac
 done

F_TERMINATE $DEBUG_MODE

# Restore stdout and stderr to terminal
F_RESET_OUTPUT
cat $LOG
exit $RC
