-- set tabstop=4 to view correctly in a vi session.

-- Description & Notes
-- -------------------
-- Generate statements to run index level REORGS.
-- So far, we've run a REORGCHK at table level only and REORG'd all tables flagged by formulae 
-- F1,F2 and F3. After RUNSTATS'ing, we've then run a second REORGCHK, this time at index level,
-- by a call to SYSPROC.REORGCHK_IX_STATS.
-- Now all we need do is REORG indexes flagged by formulae F4-F8 and RUNSTATS the tables for 
-- their index components.

-- Prioritise by smallest job first, based on the table's cardinality.

-- Substitute an alternate user temporary tablespace if required.
 
-- INPUT table $INSTANCE.reorgchk_ix   	  WORK TABLE        OUTPUT TABLE
--table_schema                            table_schema      statement (REORG,RUNSTATS,COMMIT)
--table_name                              table_name
--...
--card........................row count   priority
--...
--f4 ................ }                   f4 }
--f5 ................ }                   f5 }
--f6 ................ } formula results   f6 } booleans
--f7 ................ }                   f7 }
--f8 ................ }                   f8 }
--reorg .............. 5 character flag   reorg

-- Alex Levy, Sustainable Software Ltd. http://www.sustainablesoftware.net
-- July 2005, revised for publication on the web March 2006. 
-- May 2011,  revised for V9.5 
-- May 2012,  retested at V9.7 
-- Apr 2013,  amend logic for F4 reorgchk flag (twice!)
-- MAINTENANCE NOTES: 

--------------------------------------------------------------------------------


set current schema = current user ;

DECLARE GLOBAL TEMPORARY TABLE candidates   
(table_schema  varchar(128) not null,
 table_name    varchar(128) not null,
 priority      integer              ,
--reorg         char(5)      not null,
--f4 - f8 cannot be generated columns as the expression must not include column functions.
--f4            char(1) not null generated always as (substr(reorg,1,1)),
--f5            char(1) not null generated always as (substr(reorg,2,1)),
--f6            char(1) not null generated always as (substr(reorg,3,1)),
--f7            char(1) not null generated always as (substr(reorg,4,1)),
--f8            char(1) not null generated always as (substr(reorg,5,1))
 f4            char(1) ,
 f5            char(1) ,
 f6            char(1) ,
 f7            char(1) ,
 f8            char(1) 
) 
WITH REPLACE 
ON COMMIT DELETE ROWS 
NOT LOGGED
IN systoolstmpspace
;

-- Index REORGs work against ALL indexes on a nominated non-datapartitioned table, and tables can have more than 1 index.
-- We therefore need to group table names, such that only one REORG is carried out per table and
-- it encompasses all the recommended actions. The higher level actions include the lower level ones,
-- so a REORG INDEXES includes a CLEANUP ONLY and a CLEANUP ONLY includes CLEANUP ONLY PAGES. 
-- Note that F4 (clusterratio < 80 %) is expected for at least some indexes, if there are multiple indexes on a table;
-- therefore do not REORG all indexes for a table if only some indexes are flagged as F4 and there are no other flags for that index set.

insert into session.candidates
(table_schema, table_name, priority, f4, f5, f6, f7, f8)
select 	table_schema,
       	table_name,
       	denserank() over (order by card asc, table_name asc),
       	case when num_f4 > 0 and num_f4 = num_indexes then '*' else '-' end,
		f5, f6, f7, f8
from
       (select 	table_schema, table_name, max(indcard) as card, 
        		count(*) as num_indexes,
        		sum(case when reorg like '%*%' then 1 else 0 end) as num_flagged,
        		sum(case when reorg like '*%' then 1 else 0 end) as num_f4,
				max(substr(reorg,2,1)) as f5,
				max(substr(reorg,3,1)) as f6,
				max(substr(reorg,4,1)) as f7,
				max(substr(reorg,5,1)) as f8
        from   reorgchk_ix
        group  by table_schema,table_name) as temp
where   num_flagged > 0 
and		(num_f4 = num_indexes or f5||f6||f7||f8 like '%*%')
order by 3 asc
;

-- Now to generate REORG, RUNSTATS and STUNRATS commands:

DECLARE GLOBAL TEMPORARY TABLE statements
( statement   varchar(255) not null,
  priority    integer      not null
)
WITH REPLACE
ON COMMIT DELETE ROWS
NOT LOGGED
IN systoolstmpspace
;

-- Allow for 'short' (less than 8 characters) schema names by use of substr((length(rtrim())) -   
-- the commands permit no spaces inside a qualified table name! 
-- See maintenance note on ONLINE REORG restrictions at FP9.

-- We've already ranked the execution order for REORG statements in the priority column.
-- To interleave RUNSTATS and COMMIT statements, add a zero to each REORG rank, then add 
-- an extra 1 and 2 respectively to the priority for the RUNSTATS and COMMITs on the same table.
-- Example:
-- Table	Old Priority	New Priority	RUNSTATS priority	COMMIT priority
-- T1		1		10		11			12
-- T2		2		20		21			22

-- REORG commands
-- The Information Center gives this advice:
-- "If ... the results of the calculations for Formula 4,5 or 6 do exceed the bounds set, 
-- then index reorganization is recommended. 
-- If only the results of the calculations Formula 7 exceed the bounds set, but the results of 
-- Formula ... 4,5 and 6 are within the set bounds, then cleanup of the indexes using the 
-- CLEANUP ONLY option of REORG INDEXES is recommended. 
-- If the only calculation result to exceed the set bounds is the that of Formula 8, then a cleanup 
-- of the pseudo empty pages of the indexes using the CLEANUP ONLY PAGES option of REORG INDEXES 
-- is recommended. "

insert into session.statements     
(statement, priority)
select case
       when  f4 = '*' or f5 = '*' or f6 = '*'
       then case when indextype = 'BLOK' or num_ranges > 0
		 then 	'REORG indexes all for table '    ||
			substr(table_schema,1,length(rtrim(table_schema)))||'.'||table_name ||
                 	' allow read access;'
	       	 else 	'REORG indexes all for table '    ||
             		substr(table_schema,1,length(rtrim(table_schema)))||'.'||table_name ||
             		' allow write access;'
		 end
       when  f7 = '*'
       then 'REORG indexes all for table '    ||
             substr(table_schema,1,length(rtrim(table_schema)))||'.'||table_name ||
             ' allow write access cleanup only;'
       when  f8 = '*'
       then 'REORG indexes all for table '    ||
             substr(table_schema,1,length(rtrim(table_schema)))||'.'||table_name ||
             ' allow write access cleanup only pages;'
       else RAISE_ERROR( '73DBA','Logical error, this table needs no index REORG, Alex.')
       end,
       priority * 10
from   ( select c.table_schema, 
                c.table_name, 
                min (case when i.indextype is null or i.indextype <> 'BLOK' then 'ZZZZ' else 'BLOK' end) as indextype, 
                c.priority, 
                f4, f5, f6, f7, f8, 
                max(d.seqno) as num_ranges 
		 from  (session.candidates c inner join syscat.datapartitions d on c.table_schema = d.tabschema and c.table_name = d.tabname) 
				                     left outer join syscat.indexes i on c.table_schema = i.tabschema and c.table_name = i.tabname 
	     group by c.table_schema, c.table_name, c.priority, f4, f5, f6, f7, f8) zzz						   
;

-- RUNSTATS commands
insert into session.statements     
(statement, priority)
select 'RUNSTATS on table '||
        substr(table_schema,1,length(rtrim(table_schema)))||'.'||table_name ||
        ' for detailed indexes all allow read access;',
       (priority * 10) + 1
from   session.candidates 
;

-- COMMITs
insert into session.statements     
(statement, priority)
select 'COMMIT;',
       (priority * 10) + 2
from   session.candidates 
;

-- Create a file of interleaved REORG, RUNSTATS and COMMIT commands.
select statement from session.statements
order  by priority asc
;

-- Brevis esse Labore, obscurus fio.
