
-- Description & Notes
-- -------------------
-- 
-- Select the most appropriate indexes on which to REORG tables for a particular schema;
-- and then generate  REORG, progress query, RUNSTATS and COMMIT statements for runs during the online day.

-- A call to SYSPROC.REORGCHK_TB_STATS at schema level in a previous step populates input table 
-- reorgchk_tb and identifies tables requiring a REORG, flagged with asterisks in formulae F1/F2/F3.

-- OUTPUT COLUMNS (session.whatindexes and session.candidates)
-- tabschema  - from the input table
-- tabname    - from the input table
-- range      - name of the range partition if any
-- indtype    - index/table type governing whether INPLACE REORGs are supported
-- indschema  } identify the single most appropriate index on which to cluster the reorg;
-- indname    } select in this order of preference 
--              as NULL if there are no indexes:
--				1. MDC block dimensional indexes
--				2. clustering indexes
--				3. indexes with a time dimension column on history tables
--		  		 - will entry-sequence the data, so it becomes self-clustering;
--         		 - particularly useful in a Business Inteligence/Data Warehouse environment. 
--				4. primary constraints
--				5. unique constraints 
--				6. duplicate indexes
--      		7. nothing, if there are no indexes
-- priority   - it's smallest job first, so rank to prioritise the smallest tables first, 
--              using the CARDinality column from the input file.
--
-- This allows generation of sequenced, interleaved commands in session.statements for:
--              reorg    - The REORG comand itself for this table/index combination
--				query of the REORG's progress 
--              runstats - The RUNSTATS command for this table.
--              commit   - sic; to avoid overlong units of work.

-- Alex Levy, Sustainable Software Ltd. http://www.sustainablesoftware.net
-- July 2005, revised for publication on the web March 2006.
--
-- Modifications for V9.5:
-- 1. Use INPLACE ALLOW WRITE ACCESS, unless the table is a system catalog, MDC or range-partitioned table or has < 3 pages.
-- 	  (There are also restrictions on inplace reorgs for tables with indexes over XML columns - I have NOT catered for XML here).
-- 2. Add a query on SYSIBMADM.SNAPTAB_REORG 
-- 3. Reference control tables in the runner's implicit schema, not in schema dbadmin.
-- Alex, 05.05.2011

-- Modifications for V9.7:
-- Suppress INPLACE REORGs for any tables with XML columns or indexes.
-- Add additional range-partitioned support.
-- Retest for V9.7.
-- Alex, May 2012
--
-- MAINTENANCE NOTES: 
-- 1. These REORGs could run in the online day, but I found the INPLACE REORG to be too unwieldy at
--    V8 FP9: it runs asynchronously; it sometimes takes a while to start even when it has the database
--    to itself; it often takes a very long time to complete, even with no resource contention; 
--    when it does eventually complete, it requires a follow-up index REORG; and finally it is
--    logged. INPLACE REORGs seem better at V9.5 and V9.7, but it's your choice.
-- 2. Ancient PMR 02992,048,866 SP sysproc.reorgchk_tb_stats produces results inconsistent with 
--    the command line REORGCHK!!! In particular running a REORG against empty
--    or very small tables or tables with small rowsizes does not clear the 
--    F2 and F3 flags. Have added a (temporary) predicate to filter these out 
--    from the insert into session.candidates.
-- 3. You will need to edit the names of time dimension columns according to your own naming standards.
--    The code contains a commented out example.
-- 4. Add KEEPDICTIONARY, RESETDICTIONARY and USE TABLESPACE clauses if needed, based on your local site knowledge
-- 5. One could entrust online maintenance tasks to Optim or some Java GUI wizard,
--    but the Task Center was still in its infancy when this script was first written; 
--    my take is you have greater control with your own scripts, for example
--    in the selection of the appropriate index on which to cluster the table data.
-- 6. Edit the name of the user temporary tablespace, if you prefer not to use systooltmpspace.

-- Alex Levy, Sustainable Software Ltd. http://www.sustainablesoftware.net
--------------------------------------------------------------------------------

-- If autocommit is on, ON COMMIT DELETE ROWS will delete after EVERY sql statement
-- In the test command line environment you can either set autocommit off by
--        db2 update command options using c off; export db2options=+c   
--     or db2 -t +c -vf <filename>
-- The data will then persist until a rerun, explicit delete, commit or end of session.

set current schema = current user 
; 
DECLARE GLOBAL TEMPORARY TABLE candidates   
AS ( select table_schema                    AS tabschema,
            table_name                      AS tabname  ,
			datapartitionname				AS range,
            rank() over (order by card asc) AS priority 
     from   reorgchk_tb
   ) 
DEFINITION ONLY 
WITH REPLACE 
ON COMMIT DELETE ROWS 
NOT LOGGED
IN systoolstmpspace
;

insert into session.candidates
select table_schema,
       table_name,
	   datapartitionname,
       denserank() over (order by card asc, table_name asc)
from   reorgchk_tb
-- Accept all F1 reorgchk flags as reliable. For F2 and F3, ignore empty or miniscule tables,
-- remembering that tables are allocated  with an initial 4 data pages. 
-- Given a minimum 4k pagesize, only bother with tables having at least that much data: 
-- (max 255 rows per page on regular tablespaces could cause false flags on tables with small rowlengths).
-- FPAGES=total number of pages, NPAGES=number of pages with data on them
where  reorg like '*%'
or    (reorg like '%*%'     
and    npages > 0
and    fpages > 4
and    tsize  > 4096)
;

DECLARE GLOBAL TEMPORARY TABLE whatindex   
( tabschema   varchar(128) not null,
  tabname     varchar(128) not null,
  range       varchar(128)         ,	 
  indtype	  char(4)      not null,
  indschema   varchar(128)         ,
  indname     varchar(128)         
) 
WITH REPLACE 
ON COMMIT DELETE ROWS 
NOT LOGGED
IN systoolstmpspace
;

--------------------------------------------------------------------
-- Start of logic to select the most appropriate clustering index --
--------------------------------------------------------------------

-- MDC tables:
-- MDCs are theoretically self-clustering on the dimensional index(es). 
-- Therefore the syntax of the REORG command does not permit a specified index - the indexes are 
-- already in the system catalog. Therefore just set the index schema and name to 'MDC' pro tem.
-- Note also that a MDC table clustered on 2 dimensions will have 2 DIM indexes and 1 BLOK index;
-- a MDC with a single clustering dimension will have a DIM index; hence the distinct keyword in the
-- following select clause:

insert into session.whatindex
select distinct c.tabschema, 
       c.tabname, 
	   c.range,
       'MDC',
       'MDC',
       'MDC' 
from   session.candidates c
       INNER JOIN          
       syscat.indexes i
       ON  c.tabschema = i.tabschema         
       AND c.tabname   = i.tabname
where  i.indextype in ('DIM ','BLOK')
;

-- Clustering indexes (maximum of 1 per table).
-- The NULL predicate here and further down ensures we don't overwrite earlier index choices.
-- DB2 will automatically use the clustering index for a table reorg if the clustering index is not specified, but it's best to spell it out.
insert into session.whatindex 
select c.tabschema, 
       c.tabname, 
	   c.range,
	   'CLUS',
       i.indschema, 
       i.indname
from   (session.candidates c
       INNER JOIN          
       syscat.indexes i
       ON  c.tabschema = i.tabschema         
       AND c.tabname   = i.tabname)
       LEFT OUTER JOIN
       session.whatindex w
       ON  c.tabschema = w.tabschema         
       AND c.tabname   = w.tabname
where  i.indextype = 'CLUS'
and    w.indschema IS NULL
and    w.indname   IS NULL
;

--  Date or time dimension sequenced indexes.
--  Example is commented out below, so you can apply site naming standards.
--  Note the use of MIN to select only one index - this should select the most significant index
--  if index naming standards require declaration of the most important indexes first and
--  the index name includes an incremental sequence number.

--insert into session.whatindex 
--select c.tabschema, 
--       c.tabname, 
--	     c.range,
--		 'TIME',
--       min(i.indschema), 
--       min(i.indname)
--from   (session.candidates c
--       INNER JOIN          
--       syscat.indexes i
--       ON  c.tabschema = i.tabschema         
--       AND c.tabname   = i.tabname)
--       LEFT OUTER JOIN
--       session.whatindex w
--       ON  c.tabschema = w.tabschema         
--       AND c.tabname   = w.tabname
--where  i.colnames like '%PERIOD@_%' ESCAPE '@'
--and    w.indschema IS NULL
--and    w.indname   IS NULL
--group by c.tabschema, c.tabname
--;

-- Indexes for primary constraints.
insert into session.whatindex 
select c.tabschema, 
       c.tabname,
	   c.range,
	   'PRIM', 
       i.indschema, 
       i.indname
from   (session.candidates c
       INNER JOIN          
       syscat.indexes i
       ON  c.tabschema = i.tabschema         
       AND c.tabname   = i.tabname)
       LEFT OUTER JOIN
       session.whatindex w
       ON  c.tabschema = w.tabschema         
       AND c.tabname   = w.tabname
where  i.uniquerule = 'P'
and    w.indschema IS NULL
and    w.indname   IS NULL
;

-- Again, selecting min(indname) should return the most significant
-- of the unique indexes, if sensible naming standards have been applied!

insert into session.whatindex 
select c.tabschema, 
       c.tabname, 
	   c.range,
	   'UNIQ', 
       min(i.indschema), 
       min(i.indname)
from   (session.candidates c
       INNER JOIN          
       syscat.indexes i
       ON  c.tabschema = i.tabschema         
       AND c.tabname   = i.tabname)
       LEFT OUTER JOIN
       session.whatindex w
       ON  c.tabschema = w.tabschema         
       AND c.tabname   = w.tabname
where  i.uniquerule = 'U'
and    w.indschema IS NULL
and    w.indname   IS NULL
group by c.tabschema, c.tabname, c.range
;

-- Duplicates allowed indexes.
insert into session.whatindex 
select c.tabschema, 
       c.tabname, 
	   c.range,
	   'DUP', 
       min(i.indschema), 
       min(i.indname)
from   (session.candidates c
       INNER JOIN          
       syscat.indexes i
       ON  c.tabschema = i.tabschema         
       AND c.tabname   = i.tabname)
       LEFT OUTER JOIN
       session.whatindex w
       ON  c.tabschema = w.tabschema         
       AND c.tabname   = w.tabname
where  i.uniquerule = 'D'
and    w.indschema IS NULL
and    w.indname   IS NULL
group by c.tabschema, c.tabname, c.range
;

-- get remaining REORG candidate tables which have no indexes
insert into session.whatindex
select tabschema ,
       tabname   ,
	   range,
	   'NONE', 
       'NO INDEX',
       'NO INDEX'
from   session.candidates
       EXCEPT
select tabschema ,
       tabname   ,
	   range,
       'NONE',
	   'NO INDEX',
       'NO INDEX'
from   session.whatindex
;

-- Mark tables containing XML columns - these do not support INPLACE REORGs, but can allow write access.
update session.whatindex
set    indtype = 'XML'
where  (tabschema,tabname) in 
       ( select distinct tabschema, tabname
         from   syscat.columns
         where  typename = 'XML'
       )
;

--------------------------------------------------------------------
-- End   of logic to select the most appropriate clustering index --
--------------------------------------------------------------------

-- Now to generate REORG, RUNSTATS and COMMIT commands:

DECLARE GLOBAL TEMPORARY TABLE statements
( statement   varchar(255) not null,
  priority    integer      not null
)
WITH REPLACE
ON COMMIT DELETE ROWS
NOT LOGGED
IN systoolstmpspace
;

-- Allow for 'short' (less than 8 characters) schema names by use of substr((length(rtrim())) -   
-- the commands permit no spaces inside a qualified table name! 
-- See maintenance note on ONLINE REORG restrictions at FP9.

-- We've already ranked the execution order for REORG statements in the priority column.
-- To interleave RUNSTATS and COMMIT statements, add a zero to each REORG rank, then add 
-- an extra 1, 2 and 3 respectively to the priority for the snaptab_reorg query, RUNSTATS and COMMITs on the same table.
-- Example:
-- Table	Old Priority	New Priority	RUNSTATS priority	COMMIT priority
-- T1		1				10				12					13
-- T2		2				20				22					23

-- REORG commands
-- ALLOW WRITE ACCESS and INPLACE is not supported for MDCs; 
-- INPLACE is not supported for tables with XML columns
-- Be conservative and careful with system catalog tables! (allow read access only)
-- There are very specific rules for access on data partitioned tables and they depend on whether the indexes are partitioned; from the V9.7 Information Center:
-- " If there are no nonpartitioned indexes (except system-generated XML path indexes) defined on the table, the access mode applies only to the specified partition, 
--   users are allowed to read from and write to the other partitions of the table.
--   If there are nonpartitioned indexes defined on the table (excluding system-generated XML path indexes), the ALLOW NO ACCESS mode is the default and only supported access mode. 
--   In this case, the table is placed in ALLOW NO ACCESS mode. If ALLOW READ ACCESS is specified, SQL1548N is returned (SQLSTATE 5U047)."
-- So the simplest thing to do here is to allow the system to default the ACCESS mode for range partitioned tables.

insert into session.statements     
(statement, priority)
select 	case
	   	when w.range is null
	 	-- not a range-partitioned table
		then case
       		 when w.tabschema = 'SYSIBM'  
       			  then 	case when w.indschema is null
            				 then 'REORG table '    ||
                 				  substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname ||
                 				  ' allow read access;'
            				 else 'REORG table '    ||
                 				  substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname ||
				 				  ' index '|| 
                 				  substr(w.indschema,1,length(rtrim(w.indschema)))||'.'||w.indname ||
                 				  ' allow read access;'
						end
			when w.indtype = 'MDC'
            	 then 'REORG table '    ||
                 	   substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname ||
                 	  ' allow read access;'
       		when w.indtype = 'CLUS'
       			 then 'REORG table '    ||
    	 			  substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname || 
             	 	  ' index '||substr(w.indschema,1,length(rtrim(w.indschema)))||'.'||w.indname|| 
             		  ' inplace allow write access;'
       		when w.indtype = 'XML'
       			  then 	case when w.indschema is null
            				 then 'REORG table '    ||
                 				  substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname ||
                 				  ' allow write access;'
            				 else 'REORG table '    ||
                 				  substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname ||
				 				  ' index '|| 
                 				  substr(w.indschema,1,length(rtrim(w.indschema)))||'.'||w.indname ||
                 				  ' allow write access;'
						end
       		when w.indschema = 'NO INDEX' 
   	   			 then 'REORG table '    ||
           	 		  substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname ||
           	 		  ' inplace allow write access;'
       		else 'REORG table '    ||
             	 substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname || 
             	 ' index '||substr(w.indschema,1,length(rtrim(w.indschema)))||'.'||w.indname|| 
             	' inplace allow write access;'
       		end
	   	else
	 	-- range-partitioned table
       		case 
			when w.indschema is null
            	 then 	'REORG table '    ||
                 		substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname ||
						' on data partition ' || w.range || ';'
            	 else   'REORG table '    ||
                 		substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname ||
				 		' index '|| 
                 		substr(w.indschema,1,length(rtrim(w.indschema)))||'.'||w.indname ||
						' on data partition ' || w.range || ';'
			end
		end,	
        w.priority * 10 
from   (select w.indschema, w.indname, w.range, w.indtype, w.tabschema, w.tabname, c.priority
        from 	session.candidates c
       			INNER JOIN session.whatindex  w
       			ON  c.tabschema = w.tabschema
       			AND c.tabname   = w.tabname
	   			INNER JOIN syscat.datapartitions d	
       			ON  w.tabschema = d.tabschema
       			AND w.tabname   = d.tabname ) w
;

-- Query the REORG's progress
insert into session.statements     
(statement, priority)
select 'select reorg_status, dbpartitionnum from sysibmadm.snaptab_reorg where tabschema = '||
	   ''''||tabschema||''''||' and tabname = '||''''||tabname||''''||';', 	
       (priority * 10) + 1
from   session.candidates 
;

-- RUNSTATS commands
insert into session.statements     
(statement, priority)
select case
       when w.indschema = 'NO INDEX'
       then 'RUNSTATS on table '||
             substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname||
             ' on all columns with distribution on all columns ' || 
             'default num_freqvalues 10 num_quantiles 20 ' ||
             'allow read access;'               
       else 'RUNSTATS on table '||
             substr(w.tabschema,1,length(rtrim(w.tabschema)))||'.'||w.tabname||
             ' on all columns with distribution on all columns ' || 
             'default num_freqvalues 10 num_quantiles 20 ' ||
             'and detailed indexes all ' ||
             'allow read access;'               
       end,
       (c.priority * 10) + 2
from   session.candidates c
       INNER JOIN
       session.whatindex  w
       ON  c.tabschema = w.tabschema
       AND c.tabname   = w.tabname
;

-- COMMITs (technically not necessary since a RUNSTATS establishes its own commit point)
insert into session.statements     
(statement, priority)
select 'COMMIT;',
       (c.priority * 10) + 3
from   session.candidates c
       INNER JOIN
       session.whatindex  w
       ON  c.tabschema = w.tabschema
       AND c.tabname   = w.tabname
;

-- Create a file of interleaved REORG, progress query, RUNSTATS and COMMIT commands.
select statement from session.statements
order  by priority asc
;

-- Brevis esse Labore, obscurus fio.
